install.packages(c('shapr','xgboost','readxl'))

library(shapr)
library(xgboost)
library(readxl)

# SHAP值计算=====

## 基准结果程序=====

new_cities_data<-read_xlsx('文章使用数据.xlsx',sheet='基准指标体系')  ## 70座城市的面板数据
new_cities_data<-scale(new_cities_data[,-c(1,2)]) # 去掉日期和城市的列，保留数据列，并标准化

xg.fit<-xgboost(data=(new_cities_data[,-1]),
                label=(new_cities_data[,1]),
                nround=300,verbose = FALSE,
                params=list(eta=0.1,max_depth =12))   # XGBoost模型的拟合，相关参数由截面的滚动交叉验证得到

p<--0.6133
xg_explainer<-shapr(new_cities_data[1:(69*44),-1], xg.fit,n_combination=20000) #新建对XGBoost的SHAP解释对象 
xgboost.impact_matrix<-matrix(NA,nrow=3036,ncol=27)  # 写入的矩阵，69座城市每个有44个观测，总共27个变量（包括房价增速)

for(i in 1:(nrow(new_cities_data)/2)){
  explanation <- shapr::explain(new_cities_data[(2*i-1):(2*i),-1],approach = "empirical",  
                                explainer = xg_explainer,prediction_zero = p)               #解释两个样本中每个变量的重要性
  xgboost.impact_matrix[(2*i-1):(2*i),]<-explanation$dt[,2:(ncol(xgboost.impact_matrix)+1)] #每次写入2个样本，防止输入样本过多内存溢出  
    if(i%%10==0){   ## 每隔10步更新一遍写入程序，防止宕机引起的问题
      write.csv(xgboost.impact_matrix,'new_impact1.csv')
  }
}

colnames(xgboost.impact_matrix)<-colnames(new_cities_data)[-1]   # 将变量名写入数据中

write.csv(xgboost.impact_matrix,'XGBoost影响矩阵(基准结果).csv')   #输出最后的SHAP值矩阵


## 扩展指标体系后的结果=====
new_cities_data_1<-read_xlsx('文章使用数据.xlsx',sheet='扩展指标稳健性检验')  ## 70座城市的面板数据
new_cities_data_1<-scale(new_cities_data_1[,-c(1,2)]) # 去掉日期和城市的列，保留数据列，并标准化

xg.fit<-xgboost(data=(new_cities_data_1[,-1]),
                label=(new_cities_data_1[,1]),
                nround=300,verbose = FALSE,
                params=list(eta=0.1,max_depth =12))   

xg_explainer<-shapr(new_cities_data_1[1:(69*44),-1], xg.fit,n_combination=20000) 
xgboost.impact_matrix<-matrix(NA,nrow=3036,ncol=28)  # 扩展指标总共27个变量

for(i in 1:(nrow(new_cities_data_1)/2)){
  explanation <- shapr::explain(new_cities_data_1[(2*i-1):(2*i),-1],approach = "empirical",  
                                explainer = xg_explainer,prediction_zero = p)              
  xgboost.impact_matrix[(2*i-1):(2*i),]<-explanation$dt[,2:(ncol(xgboost.impact_matrix)+1)] #每次写入2个样本，防止输入样本过多内存溢出  
    if(i%%10==0){   ## 每隔10步更新一遍写入程序，防止宕机引起的问题
      write.csv(xgboost.impact_matrix,'new_impact1.csv')
  }
}

colnames(xgboost.impact_matrix)<-colnames(new_cities_data_1)[-1]   # 将变量名写入数据中

write.csv(xgboost.impact_matrix,'XGBoost影响矩阵(扩展稳健性检验结果).csv')   #输出最后的SHAP值矩阵

## 考虑政策滞后的结果=====
new_cities_data_2<-read_xlsx('文章使用数据.xlsx',sheet='考虑政策滞后的稳健性检验')  ## 70座城市的面板数据
new_cities_data_2<-scale(new_cities_data_2[,-c(1,2)]) # 去掉日期和城市的列，保留数据列，并标准化

xg.fit<-xgboost(data=(new_cities_data_2[,-1]),
                label=(new_cities_data_2[,1]),
                nround=300,verbose = FALSE,
                params=list(eta=0.1,max_depth =12))   

xg_explainer<-shapr(new_cities_data_2[1:(69*44),-1], xg.fit,n_combination=20000) 
xgboost.impact_matrix<-matrix(NA,nrow=3036,ncol=28)  # 扩展指标总共28个变量，最后一列为滞后的地方信贷增速

for(i in 1:(nrow(new_cities_data_2)/2)){
  explanation <- shapr::explain(new_cities_data_2[(2*i-1):(2*i),-1],approach = "empirical",  
                                explainer = xg_explainer,prediction_zero = p)              
  xgboost.impact_matrix[(2*i-1):(2*i),]<-explanation$dt[,2:(ncol(xgboost.impact_matrix)+1)] #每次写入2个样本，防止输入样本过多内存溢出  
    if(i%%10==0){   ## 每隔10步更新一遍写入程序，防止宕机引起的问题
      write.csv(xgboost.impact_matrix,'new_impact1.csv')
  }
}

colnames(xgboost.impact_matrix)<-colnames(new_cities_data_2)[-1]   # 将变量名写入数据中

write.csv(xgboost.impact_matrix,'XGBoost影响矩阵(政策滞后稳健性检验结果).csv')   #输出最后的SHAP值矩阵


## 2020年数据补充结果=====

new_cities_data_3<-read_xlsx('文章使用数据.xlsx',sheet='基准指标体系')  ## 70座城市的面板数据
new_cities_data_3<-scale(new_cities_data_3[,-c(1,2)]) 

xg.fit<-xgboost(data=(new_cities_data_3[,-1]),
                label=(new_cities_data_3[,1]),
                nround=300,verbose = FALSE,
                params=list(eta=0.1,max_depth =12))  

p<--0.6133
xg_explainer<-shapr(new_cities_data_3[1:(69*44),-1], xg.fit,n_combination=20000) 
xgboost.impact_matrix<-matrix(NA,nrow=3312,ncol=27)  # 20年数据更新后有3312个观测值

for(i in 1:(nrow(new_cities_data_3)/2)){
  explanation <- shapr::explain(new_cities_data_3[(2*i-1):(2*i),-1],approach = "empirical",  
                                explainer = xg_explainer,prediction_zero = p)               
  xgboost.impact_matrix[(2*i-1):(2*i),]<-explanation$dt[,2:(ncol(xgboost.impact_matrix)+1)]  
    if(i%%10==0){   
      write.csv(xgboost.impact_matrix,'new_impact1.csv')
  }
}

colnames(xgboost.impact_matrix)<-colnames(new_cities_data_3)[-1]   # 将变量名写入数据中

write.csv(xgboost.impact_matrix,'XGBoost影响矩阵(2020年结果).csv')   #输出最后的SHAP值矩阵

# 不同房价分化时期，影响因素贡献率的计算=====
## 设置不同分化时期计算公式=====
in_trans<-function(num,time_index){  #全样本
  new_time_index<-intersect(time_index,1:48)
  return((num-1)*48+new_time_index)
}

in_trans_1<-function(num,time_index){  #分化1期对应的样本行索引
  new_time_index<-intersect(time_index,1:16)
  return((num-1)*48+new_time_index)
}

in_trans_2<-function(num,time_index){  #分化2期对应的样本行索引
  new_time_index<-intersect(time_index,17:24)
  return((num-1)*48+new_time_index)
}

in_trans_3<-function(num,time_index){  #分化3期对应的样本行索引
  new_time_index<-intersect(time_index,25:34)
  return((num-1)*48+new_time_index)
}

in_trans_4<-function(num,time_index){  #分化4期对应的样本行索引
  new_time_index<-intersect(time_index,35:44)
  return((num-1)*48+new_time_index)
}

index_trans<-function(index,time_index){
  return(c(sapply(index,in_trans,time_index)))
}

## 分化期SHAP值的计算=====

tier_one_impact_matrix<-(as.matrix((apply((xgboost.impact_matrix[c(index_trans(1:4,which(first_class_growth>0))),]),
                      2,mean))))  #输出一线城市全样本影响因素的SHAP值排序（若计算各分化时期将“index_trans”替换为相应时期的索引函数即可）

tier_two_impact_matrix<-(as.matrix((apply((xgboost.impact_matrix[c(index_trans(5:35,which(first_class_growth>0))),]),
                      2,mean))))  #输出二线城市全样本影响因素的SHAP值排序

tier_three_impact_matrix<-(as.matrix((apply((xgboost.impact_matrix[c(index_trans(36:65,which(first_class_growth>0))),]),
                      2,mean))))  #输出三线城市全样本影响因素的SHAP值排序
